/* Assembler for comp 3 
 * Michael Robinson
 * 17 December 2001: Finished!
 * 10 November 2001: Initial Revision
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

int assemble_file( char *infile, char *ofile );
int opcodeout( char *opcode );

int add_symbol( char *symbol, int address );
int find_symbol( char *symbol );
void kill_symboltable( void );
int create_symboltable( void );

int nsym_alloc;
int nsym;
char **sym;
int *s_add;

main( int argc, char *argv[] )
{
   char infile[80], ofile[80];

   /* Parse command line */
   if( argc < 2 || argc > 3 )
   {
      fprintf(stderr,"usage: assembler infile [outfile]\n" );
      exit(-1);
   }
   strcpy( infile, argv[1] );
   if( argc >= 3 )
      strcpy( ofile, argv[2] );
   else
      strcpy( ofile, infile );

   if( create_symboltable() != 0 )
     {
       fprintf(stderr,"FATAL: Couldn't allocate symbol table!\n");
       exit(-1);
     }

   assemble_file( infile, ofile );

   kill_symboltable();
}

/* Parse and assemble files
 * Input: infile = assembly file to read
 *        ofile  = prefix for output files
 * Output: none
 * Returns: 0  = success
 *          -1 = file open error
 * Note: produces the following output files:
 *       ofile_rom1.dat, ofile_rom2.dat, ofile_ram.dat          
 */
int assemble_file( char *infile, char *ofile )
{
   FILE *fpin, *fpr1, *fpr2;
   char str[80], line[80], *jnk, opcode[5], cond2[4];
   int cstate, jtable[16], i, jmp, j, addr, dest, line_num;

   /* Open files */
   if( ( fpin = fopen( infile, "r" ) ) == NULL )
   {
      fprintf(stderr,"FATAL: Can't open input: %s\n", infile );
      return -1;
   }
   strcpy( str, ofile );
   strcat( str, "_rom1.dat");
   if( ( fpr1 = fopen( str, "w" ) ) == NULL )
   {
      fprintf(stderr,"FATAL: Can't open output: %s\n", str );
      fclose( fpin );
      return -1;
   }
   strcpy( str, ofile );
   strcat( str, "_rom2.dat");
   if( ( fpr2 = fopen( str, "w" ) ) == NULL )
   {
      fprintf(stderr,"FATAL: Can't open output: %s\n", str );
      fclose( fpin );
      fclose( fpr1 );
      return -1;
   }

   /* Operation: Two passes 
    * First pass: Symbol collection
    * Second pass: Actual assembly phase 
    */

   /* First pass */
   cstate = 0;
   line_num = 0;
   while( fgets( line, 80, fpin ) != NULL )
   {
      line_num ++;
      /* Kill off comments and newlines */
      jnk = strstr( line, ";" );
      if( jnk != NULL )
         *jnk = '\0';
      jnk = strstr( line, "\n" );
      if( jnk != NULL )
         *jnk = '\0';

      /* Prepare jumptable */
      for( i = 0; i < 16; i ++ )
         jtable[i] = cstate + 1;

      /* Remove whitespace */
      for( i = 0; isspace( line[i] ); i ++ );
      if( line[i] == '\0' )
         continue;

      /* Parse off opcode */
      for( j = 0; !isspace(line[i]) && line[i] != '\0'; j ++, i ++ )
         opcode[j] = line[i];
      opcode[j] = '\0';
      if( opcode[j-1] != ':' ) /* Found a non-symbol: next line! */
	{
	  cstate ++;
	  continue;
	}

      opcode[j-1] = '\0';
      if( add_symbol( opcode, cstate ) < 0 )
	{
	  fprintf(stderr,"FATAL: %s(%d): Couldn't allocate symbol!\n",
		  infile, line_num);
	  goto fatality;
	}
   }
   
   /* Second pass */
   rewind( fpin );
   cstate = 0;
   line_num = 0;
   while( fgets( line, 80, fpin ) != NULL )
   {
      line_num ++;
      /* Kill off comments and newlines */
      jnk = strstr( line, ";" );
      if( jnk != NULL )
         *jnk = '\0';
      jnk = strstr( line, "\n" );
      if( jnk != NULL )
         *jnk = '\0';

      /* Prepare jumptable */
      for( i = 0; i < 16; i ++ )
         jtable[i] = cstate + 1;

      /* Remove whitespace */
      for( i = 0; isspace( line[i] ); i ++ );
      if( line[i] == '\0' )
         continue;

      /* Parse off opcode */
      for( j = 0; !isspace( line[i] ) && line[i] != '\0'; j ++, i ++ )
         opcode[j] = line[i];
      opcode[j] = '\0';
      if( opcode[j-1] == ':' ) /* Found a symbol not an opcode: next line! */
	continue;

      if( line[i] == '\0' )
      {
         fprintf(stderr,"ERROR: %s(%d): Must include address and destination\n",
                 infile, line_num);
         continue;
      }

      /* Ignore whitespace */
      for( ; isspace( line[i] ); i ++ );
      if( line[i] == '\0' )
	 continue;

      /* Parse off address */
      for( j = 0; !isspace( line[i] ) && line[i] != '\0'; j ++, i ++ )
         str[j] = line[i];
      str[j] = '\0';
      if( line[i] == '\0' )
      {
         fprintf(stderr,"ERROR: %s(%d): Must include destination\n",
                 infile, line_num);
         continue;
      }
      if( sscanf(str, "%x", &addr ) != 1 )
      {
         fprintf(stderr,"ERROR: %s(%d): Address not valid\n",
                 infile, line_num);
         continue;
      }

      /* Ignore whitespace */
      for( ; isspace( line[i] ); i ++ );
      if( line[i] == '\0' )
	 continue;

      /* Parse off destination */
      for( j = 0; !isspace( line[i] ) && line[i] != '\0'; j ++, i ++ )
         str[j] = line[i];
      str[j] = '\0';
      if( tolower(str[0]) == 'a' || str[0] == '0' )
         dest = 0;
      else
         dest = 1;

      /* Look for modifiers to the jumptable */
      while( line[i] != '\0' )
	{
	  /* Ignore separating space */
	  for( ; isspace( line[i] ); i ++ );
	  if( line[i] == '\0' )
	    break;
	  
	  /* Parse each modifier */
	  for( j = 0; !isspace( line[i] ) && line[i] != '\0'; j ++, i ++ )
            str[j] = line[i];
	  str[j] = '\0';
	  if( str[0] == '\0' ) /* Comments */
	    break;
	 	  
	  jnk = strchr( str, ':' );
	  if( jnk == NULL )
	    {
	      fprintf(stderr,"WARNING: %s(%d): Invalid modifier.\n", 
		      infile, line_num);
	      break;
	    }
	  
	  for( jmp = 0; jmp < 16; jmp ++ )
	    {
	      cond2[0] = tolower( str[0] );
	      cond2[1] = tolower( str[1] );
	      cond2[2] = tolower( str[2] );
	      cond2[3] = tolower( str[3] );
	      if( ( ( ( jmp & 0x01 ) && 
		      ( ( cond2[3] == '1' ) || ( cond2[3] == 'x') ) ) ||
		    ( !( jmp & 0x01 ) && 
		      ( ( cond2[3] == '0' ) || ( cond2[3] == 'x') ) ) ) &&
		  ( ( ( jmp & 0x02 ) && 
		      ( ( cond2[2] == '1' ) || ( cond2[2] == 'x') ) ) ||
		    ( !( jmp & 0x02 ) && 
		      ( ( cond2[2] == '0' ) || ( cond2[2] == 'x') ) ) ) &&
		  ( ( ( jmp & 0x04 ) && 
		      ( ( cond2[1] == '1' ) || ( cond2[1] == 'x') ) ) ||
		    ( !( jmp & 0x04 ) && 
		      ( ( cond2[1] == '0' ) || ( cond2[1] == 'x') ) ) ) &&
		  ( ( ( jmp & 0x08 ) && 
		      ( ( cond2[0] == '1' ) || ( cond2[0] == 'x') ) ) ||
		    ( !( jmp & 0x08 ) && 
		      ( ( cond2[0] == '0' ) || ( cond2[0] == 'x') ) ) ) )
		{
		  j = find_symbol( jnk + 1 );
		  if( j < 0 )
		    {
		      fprintf(stderr,"ERROR: %s(%d): Undefined symbol: %s\n",
			      infile, line_num, jnk + 1);
		      break;
		    }
		  jtable[jmp] = j;
		}
	    }
	}
      
      /* Convert opcode mnemonic->numerical value */
      j = opcodeout( opcode );
      if( j < 0 )
	{
	  fprintf(stderr,"ERROR: %s(%d): Opcode not valid\n",
		  infile, line_num);
	  continue;
	}
      j = (j << 1) | (dest & 0x01);

      /* Now we have a complete instruction, write it to output files */
      for( i = 0; i < 16; i ++ )
	fprintf( fpr1, "%x\n", jtable[i] );
      fprintf(fpr1,"\n");
      
      fprintf( fpr2, "@%x\n%x%x\n",cstate*16,addr, j );
      cstate ++;
   } 

 fatality:

   /* Close files */
   fclose( fpin );
   fclose( fpr1 );
   fclose( fpr2 );
}

/* Convert a Mnemonic opcode into a numerical one
 * Input: opcde = string of opcode
 * Returns: Opcode number or -1 if invalid opcode 
 */
int opcodeout( char *opcode )
{
  char opcde[4];

  opcde[0] = toupper( opcode[0] );
  opcde[1] = toupper( opcode[1] );
  opcde[2] = toupper( opcode[2] );
  opcde[3] = '\0';

  if( !strcmp( opcde, "ADD" ) )
    return 0;
  if( !strcmp( opcde, "CMP" ) )
    return 1;
  if( !strcmp( opcde, "ANM" ) )
    return 2;
  if( !strcmp( opcde, "AND" ) )
    return 3;
  if( !strcmp( opcde, "ORP" ) )
    return 4;
  if( !strcmp( opcde, "ORN" ) )
    return 5;
  if( !strcmp( opcde, "DEC" ) )
    return 6;
  if( !strcmp( opcde, "ACC" ) )
    return 7;

  return -1;
}

/* Add a symbol to the symbol table
 * Input: symbol  = Symbol name
 *        address = Symbol address
 * Returns: -1 = Couldn't allocate more symbols
 *          -2 = Couldn't allocate data for symbol
 *          otherwise returns the number of symbols in the table
 */
int add_symbol( char *symbol, int address )
{
  int i;

  if( nsym == nsym_alloc )
    {
      sym = (char **)realloc(sym, nsym_alloc * 2 * sizeof( char *));
      if( sym == NULL )
	return -1;

      s_add = (int *)realloc(s_add, nsym_alloc * 2 * sizeof( int ));
      if( s_add == NULL )
	return -1;

      nsym_alloc *= 2;
    }
  sym[nsym] = (char *)malloc( ( strlen( symbol ) + 1 ) * sizeof( char ) );
  if( sym[nsym] == NULL )
    return -2;

  for( i = 0; i < strlen( symbol ); i ++ )
    sym[nsym][i] = tolower( symbol[i] );
  s_add[nsym] = address;

  nsym ++;

  return nsym;
}

/* Find symbol in table
 * Input: symbol = Symbol to search for
 * Returns: -1 = symbol not found
 *          else the address of the symbol
 */
int find_symbol( char *symbol )
{
  int i, j;
  char f;

  for( i = 0; i < nsym; i ++ )
    {
      if( strlen( sym[i] ) != strlen( symbol ) )
	  continue;

      f = 0;
      for( j = 0; j < strlen( symbol ); j ++ )
	{
	  if( sym[i][j] != tolower( symbol[j] ) )
	    {
	      f = 1;
	      break;
	    }
	}
      if( !f )
	return s_add[i];
    }

  return -1;
}

/* Free the symbol table */
void kill_symboltable( void )
{
  int i;

  if( s_add && ( nsym_alloc > 0 ) )
    {
      free( s_add );
      s_add = NULL;
    }

  if( sym && ( nsym_alloc > 0 ) )
    {
      for( i = 0; i < nsym; i ++ )
	{
	  if( sym[i] != NULL )
	    {
	      free( sym[i] ); 
	      sym[i] == NULL;
	    }
	}
      
      nsym = 0;
      nsym_alloc = 0;
      free( sym );
    }
}

/* Allocate a nominal symbol table
 * Returns: 0 = success
 *         -1 = couldn't allocate table
 */
int create_symboltable( void )
{
  nsym = 0;
  nsym_alloc = 0;
  
  sym = (char **)malloc( 10 * sizeof( char *));
  if( sym == NULL )
    {
      return -1;
    }

  s_add = (int *)malloc( 10 * sizeof( int ) );
  if( s_add == NULL )
    {
      free( sym );
      sym = NULL;
      return -1;
    }

  nsym_alloc = 10;

  return 0;
}

